//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#define mapunitCPP
#include "mapunit.h"
#include "xFiles.hpp"

#include <typeinfo>

//---------------------------------------------------------------------------

#pragma package(smart_init)


// default constructor
TCell::TCell()
{
  init(0, 0, 0);

  FBreakPtr = NULL;
}

void TCell::init(int Layer, int x, int y)
{
  FLayer = Layer;
  FX = x;
  FY = y;

  // Ω}HϮKϰʧ@ bitmap
  FCellBitmap = new Graphics::TBitmap;
  FCellBitmap->Width = TILE_WIDTH;
  FCellBitmap->Height = TILE_HEIGHT;
  FCellBitmap->Transparent = true;
  FCellBitmap->TransparentColor = TRANSPARENT_COLOR;

  FSMTileBitmap = new Graphics::TBitmap;
  FSMTileBitmap->Width = SM_TILE_WIDTH;
  FSMTileBitmap->Height = SM_TILE_HEIGHT;
  FSMTileBitmap->Transparent = true;
  FSMTileBitmap->TransparentColor = TRANSPARENT_COLOR;
}

TCell::~TCell()
{
  DisposeBreakMap();
  
  delete FSMTileBitmap;
  delete FCellBitmap;
}

// ھڮyФί}HpҦΰϰ
void TCell::CalcRect()
{
  if (!FBreakPtr) // YS}H, hΰϰܦn
    FRect = Classes::Rect(TILEWIDTH[FX], TILEHEIGHT[FY], TILEWIDTH[FX + 1], TILEHEIGHT[FY + 1]);
  else {
    TBreakMap& BreakMap = *FBreakPtr;

    // Y}H, hΰϰnwWUkOo
    FRect = Classes::Rect(0, 0, 0, 0);

    // ̥
    for (int x = 0; x < SM_TILE_NUM_X; x++)
      for (int y = 0; y < SM_TILE_NUM_Y; y++)
        if (!BreakMap[x][y]) {
          FRect.Left = TILEWIDTH[FX] + SM_TILEWIDTH[x];
          goto OutLoop1;
        }

    OutLoop1:

      // ̥k
    for (int x = SM_TILE_NUM_X - 1; x >= 0; x--)
      for (int y = 0; y < SM_TILE_NUM_Y; y++)
        if (!BreakMap[x][y]) {
          FRect.Right = TILEWIDTH[FX] + SM_TILEWIDTH[x + 1];
          goto OutLoop2;
        }

    OutLoop2:

      // ̤W
    for (int y = 0; y < SM_TILE_NUM_Y; y++)
      for (int x = 0; x < SM_TILE_NUM_X; x++)
        if (!BreakMap[x][y]) {
          FRect.Top = TILEHEIGHT[FY] + SM_TILEHEIGHT[y];
          goto OutLoop3;
        }

    OutLoop3:

      // ̤U
    for (int y = SM_TILE_NUM_Y - 1; y >= 0; y--)
      for (int x = 0; x < SM_TILE_NUM_X; x++)
        if (!BreakMap[x][y]) {
          FRect.Bottom = TILEHEIGHT[FY] + SM_TILEHEIGHT[y + 1];
          goto OutLoop4;
        }

    OutLoop4:

    // }
    if (FRect == Classes::Rect(0, 0, 0, 0)) {
      FTileNo = 0; // oӹϮ}, F讳
    }
  }
}

// qƬyŪϮ
void TCell::LoadFromStream(TStream* Stream)
{
  bool bHasBreakPtr;

  TReader* reader = new TReader(Stream, 2048);
  try {
    FTileNo = reader->ReadInteger();

    bHasBreakPtr = reader->ReadBoolean();

    if (bHasBreakPtr) { // Y}H, hŪ
      AllocBreakMap();
      reader->Read(FBreakPtr, sizeof(TBreakMap));
    } else
      DisposeBreakMap();
  } __finally {
    delete reader;
  }

  BreakMapChanged();
}

void TCell::BreakMapChanged()
{
  CalcRect(); // spҦΰϰ
  BuildCellBitmap(); // إ߹Ϯϧ (iO}H)
}

// NϮgJƬy
void TCell::SaveToStream(TStream* Stream)
{
  TWriter* writer = new TWriter(Stream, 2048);
  try {
    writer->WriteInteger(FTileNo);

    bool bHasBreakPtr = FBreakPtr != NULL; // O_֦}H
    writer->WriteBoolean(bHasBreakPtr);
    if (bHasBreakPtr) // Y}H, hgJ
      writer->Write(FBreakPtr, sizeof(TBreakMap));
  } __finally {
    delete writer;
  }
}

TTileAttr TCell::GetTileAttr()
{
  return TTiles::Instance().Tile[FTileNo].Attr;
}

void TCell::SetTileAttr(TTileAttr Value)
{
  TTiles::Instance().Tile[FTileNo].Attr = Value;
}

// omϮ檺Ϥ
Graphics::TBitmap* TCell::GetTileBitmap()
{
  return TTiles::Instance().Tile[FTileNo].Bitmap;
}

// إ߹Ϯ}H
void TCell::AllocBreakMap()
{
//  if (!FBreakPtr) FBreakPtr = new TBreakMap;
  if (!FBreakPtr) FBreakPtr = (PBreakMap)malloc(sizeof(TBreakMap));

  memset(FBreakPtr, 0, sizeof(TBreakMap));
}

// Ϯ}H
void TCell::DisposeBreakMap()
{
  if (FBreakPtr) {
    // delete[] FBreakPtr;
    free (FBreakPtr);
    FBreakPtr = NULL;

    BreakMapChanged();
  }
}

// ̯xΰϰ]wϮ}H
void TCell::BreakBy(TRect ARect)
{
  // Y٨S}H, Nؤ@
  if (!FBreakPtr) AllocBreakMap();

  TRect R, R1;

  // P ARect 涰, 涰ph]}
  for (int y = 0; y < SM_TILE_NUM_Y; y++)
    for (int x = 0; x < SM_TILE_NUM_X; x++) {
      // pp檺xΰϰ
      R1 = SM_TILE_RECT;
      OffsetRect(&R1, SM_TILEWIDTH[x], SM_TILEHEIGHT[y]);
      OffsetRect(&R1, TILEWIDTH[FX], TILEHEIGHT[FY]);

      // Y涰, h}
      if (IntersectRect(&R, &ARect, &R1))
        (*FBreakPtr)[x][y] = true;
    }

  BreakMapChanged();
}

TCell& TCell::operator=(const TCell& c)
{
  if (this != &c) { // קKۧڳ]
    FTileNo = c.FTileNo;

    FLayer = c.FLayer;
    FX = c.FX;
    FY = c.FY;
    
    FRect = c.FRect;

    // }a]ƻsL
    if (c.FBreakPtr) {
      AllocBreakMap();
      memcpy(FBreakPtr, c.FBreakPtr, sizeof(TBreakMap));
    } else
      DisposeBreakMap();
  }
  return *this;
}

bool TCell::IsBroken(int x, int y)
{
  return (*FBreakPtr)[x][y];
}

// ϮO_qL
bool TCell::GetCanPass()
{
  return (TileNo == 0 || TileAttr.Contains(taCanPass));
}

void TCell::SetTileNo(int Value)
{
  if (FTileNo != Value) {
    FTileNo = Value;
    BreakMapChanged(); // sإ߹Ϯϧ
  }
}

// Nwaϼheb Canvas W
void TCell::Draw(TCanvas* Canvas)
{
  if (FLayer != LAYER_TERR) { // Daμh
    if (FTileNo == 0) return; // S]w~

    Canvas->Draw(TILEWIDTH[FX], TILEHEIGHT[FY], FCellBitmap);
  } else {
     // aμhݭnzK
     BitBlt(Canvas->Handle, TILEWIDTH[FX], TILEHEIGHT[FY],
       TILE_WIDTH, TILE_HEIGHT, FCellBitmap->Canvas->Handle, 0, 0, SRCCOPY);
  }
}

// øs}HpϮ
void TCell::BuildCellBitmap()
{
  if (!FBreakPtr) {
    FCellBitmap->Assign(GetTileBitmap());
    return;
  }
  
  //  FTileBitmap ]z
  FCellBitmap->Canvas->Brush->Color = TRANSPARENT_COLOR;
  FCellBitmap->Canvas->Brush->Style = bsSolid;
  FCellBitmap->Canvas->FillRect(TILE_RECT);

  // Y}HϮ, h@@NsbpϮKW
  for (int n = 0; n <= 3; n++)
    for (int m = 0; m <= 3; m++) {

      // YpϮ}FNεe
      if (IsBroken(m, n)) continue;

      // ƻst@ bitmap
      FSMTileBitmap->Canvas->CopyRect(
        SM_TILE_RECT,
        GetTileBitmap()->Canvas,
        Classes::Rect(SM_TILEWIDTH[m], SM_TILEHEIGHT[n],
          SM_TILEWIDTH[m + 1], SM_TILEHEIGHT[n + 1]));

      // AKeW, HFzKϮĪG
      FCellBitmap->Canvas->Draw(SM_TILEWIDTH[m], SM_TILEHEIGHT[n],
        FSMTileBitmap);
    }
}

// TMap class

TMap* TMap::FInstance = NULL;

TMap& TMap::Instance()
{
  if (FInstance == NULL) {
    FInstance = new TMap;
  }
  return *FInstance;
}

TMap::TMap()
{
  FLevelNo = 0;

  // ]wC@hҦϮ
  for (int l = 0; l <= LAYER_MAX; l++)
    for (int y = 0; y < TILE_NUM_Y; y++)
      for (int x = 0; x < TILE_NUM_X; x++)
        FMaps[l][x][y].init(l, x, y);
}

TMap::~TMap()
{
}

AnsiString TMap::GetFileName()
{
  return Format("%s%s%d%s", OPENARRAY(TVarRec, (AppDir, FN_MAP_PREFIX, FLevelNo, FN_MAP_EXT)));
}

// qɮŪa
void TMap::LoadFromFile()
{
  TFileStream* fs = new TFileStream(GetFileName(), fmOpenRead);
  TReader* reader = new TReader(fs, 2048);
  try {
    CheckSignature(reader, SIG_MYFILE); // ˬdY
    CheckSignature(reader, typeid(*this).name()); // ˬdY

    // Ūlm
    FRole_X = reader->ReadInteger();
    FRole_Y = reader->ReadInteger();

    reader->FlushBuffer();

    // ŪC@hҦϮ
    for (int i = 0; i <= LAYER_MAX; i++)
      for (int y = 0; y < TILE_NUM_Y; y++)
        for (int x = 0; x < TILE_NUM_X; x++)
          FMaps[i][x][y].LoadFromStream(fs);
  } __finally {
    delete reader;
    delete fs;
  }
}

// NaϼgJɮ
void TMap::SaveToFile()
{
  TFileStream* fs = new TFileStream(GetFileName(), fmCreate | fmOpenWrite);
  TWriter* writer = new TWriter(fs, 2048);
  try {
    writer->WriteString(SIG_MYFILE);
    writer->WriteString(typeid(*this).name());

      // gJlm
    writer->WriteInteger(FRole_X);
    writer->WriteInteger(FRole_Y);

    writer->FlushBuffer();

    // gJC@hҦϮ
    for (int i = 0; i <= LAYER_MAX; i++)
      for (int y = 0; y < TILE_NUM_Y; y++)
        for (int x = 0; x < TILE_NUM_X; x++)
          FMaps[i][x][y].SaveToStream(fs);
  } __finally {
    delete writer;
    delete fs;
  }
}

void TMap::SetLevelNo(int Value)
{
  FLevelNo = Value;
  LoadFromFile();
}


// Nwaϼheb Canvas W
void TMap::Draw(TCanvas* Canvas, int Layer)
{
  for (int y = 0; y < TILE_NUM_Y; y++) // C@Ϯ
    for (int x = 0; x < TILE_NUM_X; x++)
      FMaps[Layer][x][y].Draw(Canvas);
}

// ]a
void TMap::ResetAllLayers()
{
  for (int i = 0; i <= LAYER_MAX; i++)
    ResetLayer(i);
  
  FRole_X = 0;
  FRole_Y = 0;
}

// ]waϼh
void TMap::ResetLayer(int Layer)
{
  for (int y = 0; y < TILE_NUM_Y; y++)
    for (int x = 0; x < TILE_NUM_X; x++)
      FMaps[Layer][x][y].TileNo = 0;
}

TCell& TMap::GetCell(int Layer, int x, int y)
{
  return FMaps[Layer][x][y];
}

void TMap::SetRole_X(int Value)
{
  FRole_X = Value;
}

void TMap::SetRole_Y(int Value)
{
  FRole_Y = Value;
}

