//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <mmsystem.h>

#define spriteCPP
#include "sprite.h"
#include "xFiles.hpp"
#include "xKernel.hpp"
#include "mapunit.h"
//---------------------------------------------------------------------------

#pragma package(smart_init)
TSprite::TSprite()
{
  // l
  FX = 0;
  FY = 0;
  FFrameNo = 0;

  FMoveDelay = 0;
  FMoveDelayCount = 0;

  FActive = true;
  FVisible = true;
  FPostToDead = false;
  FOnAir = false;

  FAttr = TSpriteAttr();
  FAdvanceFrameMode = afWrap;
  FSpeed = SPRITE_DEFAULT_SPEED;

  // w]ؤojp
  FObjectRect = TILE_RECT;

  // ⪺ bitmap
  FBits = new Graphics::TBitmap;

  // zKϥΪ bitmap
  FInvBitmap = new Graphics::TBitmap;
  FInvBitmap->Transparent = true;
  FInvBitmap->TransparentColor = TRANSPARENT_COLOR;
}

TSprite::~TSprite()
{
  delete FInvBitmap;
  delete FBits;
}

void TSprite::LoadBits()
{
  // Ūv
  FBits->LoadFromFile(AppDir + DR_IMAGES + FFileName);

  // ]wzKϥΪ bitmap ؤo
  FInvBitmap->Width = ObjectWidth;
  FInvBitmap->Height = ObjectHeight;

  // ˬdJ bitmap ؤoO_Xk
  if (!Attr.Contains(saUndirectionalBitmap))
    if (FBits->Width < ObjectWidth * 4) // |ӤV
      throw Exception("Width of bitmap is invalid");

  if (FBits->Height < ObjectHeight * (FFrameMax + 1))
    throw Exception("Height of bitmap is invalid");
}

void TSprite::ResetStatus() // ]⪬A
{
  FPostToDead = false;
}

void TSprite::Draw(TCanvas* Canvas)
{
  //YiΤwnORhe
  if (!FVisible || FPostToDead) return;

  // ʵeؽsXkhe
  if (FFrameNo > FFrameMax || FFrameNo < 0) return;

  // NnqXϰ FInvBitmap
  // Yu@iϧ, V
  if (Attr.Contains(saUndirectionalBitmap))
    FInvBitmap->Canvas->CopyRect(ObjectRect, FBits->Canvas,
    Classes::Rect(0, FFrameNo * ObjectHeight, ObjectWidth,
    (FFrameNo + 1) * ObjectHeight));
  else  // ̥ثeV
    FInvBitmap->Canvas->CopyRect(ObjectRect, FBits->Canvas,
      Classes::Rect((int)FDirection * ObjectWidth,
      FFrameNo * ObjectHeight, ((int)FDirection + 1) * ObjectWidth,
      (FFrameNo + 1) * ObjectHeight));

  // AeX FInvBitmap
  Canvas->Draw(FX, FY, FInvBitmap);
}

bool TSprite::CheckBoundCollisions(int& x, int& y)
{
  int OrgX, OrgY;

  // ˬd
  OrgX = x; OrgY = y;

  if (y < 0) y = 0; // O_WXW
  if (x < 0) x = 0; // O_WX
  if (x + ObjectWidth >= WORLD_WIDTH) // O_WXk
    x = WORLD_WIDTH - ObjectWidth;
  if (y + ObjectHeight >= WORLD_HEIGHT) // O_WXU
    y = WORLD_HEIGHT - ObjectHeight;

  // O_ ?
  return (OrgX != x || OrgY != y);
}

bool TSprite::IsRealCellCollision(int Layer, int CellXPos, int CellYPos, int& x, int& y)
{
  bool bCollision = false;

  TMap& Map = TMap::Instance();
  TCell& Cell = Map.GetCell(Layer, CellXPos, CellYPos);

  // P_O_uIaΪ
  switch (FDirection) {
    case drUp:
      if (CellYPos != Tile_Y + 1) { // VWɤi]IUa
        y = Cell.Rect.Bottom;
        bCollision = true;
      }
      break;

    case drDown:
      if (CellYPos != Tile_Y - 1) {
        y = Cell.Rect.Top - ObjectHeight;
        bCollision = true;
      }
      break;

    case drLeft:
      if (CellXPos != Tile_X + 1) {
        x = Cell.Rect.Right;
        bCollision = true;
      }
      break;

    case drRight:
      if (CellXPos != Tile_X - 1) {
        x = Cell.Rect.Left - ObjectWidth;
        bCollision = true;
      }
      break;
  }

  // ʮɦ۰ʹaΪ, _Ӥ

  if (bCollision && FAttr.Contains(saAlignWithTerrItem))
    switch (FDirection) {
      case drUp: case drDown:
        if (CellXPos == Tile_X && abs(x + ObjectWidth - Map.GetCell(Layer, CellXPos, CellYPos).Rect.Left) <= SMOOTH_MOVE_THRESHOLD)
          x = Map.GetCell(Layer, CellXPos, CellYPos).Rect.Left - ObjectWidth;
        else if (CellXPos == Tile_X && abs(x - Map.GetCell(Layer, CellXPos, CellYPos).Rect.Right) <= SMOOTH_MOVE_THRESHOLD)
          x = Map.GetCell(Layer, CellXPos, CellYPos).Rect.Right;
        else if (CellXPos == Tile_X - 1 && abs(x - TILEWIDTH[Tile_X]) <= SMOOTH_MOVE_THRESHOLD && Map.GetCell(Layer, Tile_X, Tile_Y).CanPass)
          x = TILEWIDTH[Tile_X];
        else if (CellXPos == Tile_X + 1 && abs(x + ObjectWidth - TILEWIDTH[CellXPos]) <= SMOOTH_MOVE_THRESHOLD && Map.GetCell(Layer, CellXPos - 1, Tile_Y).CanPass)
          x = TILEWIDTH[CellXPos] - ObjectWidth;
        break;

      case drLeft: case drRight:
        if (CellYPos == Tile_Y && abs(y + ObjectHeight - Map.GetCell(Layer, CellXPos, CellYPos).Rect.Top) <= SMOOTH_MOVE_THRESHOLD)
          y = Map.GetCell(Layer, CellXPos, CellYPos).Rect.Top - ObjectHeight;
        else if (CellYPos == Tile_Y && abs(y - Map.GetCell(Layer, CellXPos, CellYPos).Rect.Bottom) <= SMOOTH_MOVE_THRESHOLD)
          y = Map.GetCell(Layer, CellXPos, CellYPos).Rect.Bottom;
        else if (CellYPos == Tile_Y - 1 && abs(y - TILEHEIGHT[Tile_Y]) <= SMOOTH_MOVE_THRESHOLD && Map.GetCell(Layer, Tile_X, Tile_Y).CanPass)
          y = TILEHEIGHT[Tile_Y];
        else if (CellYPos == Tile_Y + 1 && abs(y + ObjectHeight - TILEHEIGHT[CellYPos]) <= SMOOTH_MOVE_THRESHOLD && Map.GetCell(Layer, Tile_X, CellYPos - 1).CanPass)
          y = TILEHEIGHT[CellYPos] - ObjectHeight;
      break;
    }

  return bCollision;
}

void TSprite::CheckCellCollisionsPos(TCellArray& Collisions, const TRect& SpriteRect, int Layer, int CellXPos, int CellYPos, int& x, int& y)
{
  if (CellXPos < 0 || CellYPos < 0 || CellXPos >= TILE_NUM_X || CellYPos >= TILE_NUM_Y) return;

  TMap& Map = TMap::Instance();
  TCell& Cell = Map.GetCell(Layer, CellXPos, CellYPos);

  // NqLa Cell [JIG}C
  TRect R, R1 = Cell.Rect;
  if (!Cell.CanPass &&
    IntersectRect(&R, &SpriteRect, &R1) &&
    IsRealCellCollision(Layer, CellXPos, CellYPos, x, y)) {

    Collisions.push_back(&Cell);
  }
}

bool TSprite::CheckCellCollisions(int Layer, int& x, int& y, TCellArray& Collisions)
{
  TRect SpriteRect = ObjectRect;
  OffsetRect(&SpriteRect, x, y);

  int CellXPos = (x + ObjectWidth / 2) / TILE_WIDTH;
  int CellYPos = (y + ObjectHeight / 2) / TILE_HEIGHT;

  // OըPKϮΥҦbϮ, O_o͸I
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos - 1, CellYPos - 1, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos, CellYPos - 1, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos + 1, CellYPos - 1, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos + 1, CellYPos, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos, CellYPos, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos - 1, CellYPos, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos - 1, CellYPos + 1, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos, CellYPos + 1, x, y);
  CheckCellCollisionsPos(Collisions, SpriteRect, Layer, CellXPos + 1, CellYPos + 1, x, y);

  return Collisions.size() > 0; // ܦIo
}

bool TSprite::CheckSpriteCollisions(TSpriteArray& Sprites, int& x, int& y, TSpriteArray& Collisions)
{
  TRect SpriteRect, R;

  // psxΰϰ
  SpriteRect = ObjectRect;
  OffsetRect(&SpriteRect, x, y);

  // @@MXǤJ
  for (TSpriteArray::iterator p = Sprites.begin(); p != Sprites.end(); p++) {
    TSprite* &S = *p;
    TRect SRect = S->Rect;

    if (this != S && !S->PostToDead && S->Visible &&
      IntersectRect(&R, &SpriteRect, &SRect)) {

      // N쪺[JIG}C
      Collisions.push_back(S);
    }
  }

  return Collisions.size() > 0; // ܦIo
}

// (X, y) OϥΪsy, Ǧ^ȪܬO_oͥI
bool TSprite::CheckCollisions(int& x, int& y)
{
  if (FPostToDead) return false; // wgnOǳƳQRF

  bool bCollision = false;
  
  // ˬdO_aΪ
  if (!bCollision && !FAttr.Contains(saNoCellCollision)) {
    TCellArray CellCollisions;
    bCollision = CheckCellCollisions(LAYER_TERRITEM, x, y, CellCollisions);
  }

  // ˬdO_
  if (!bCollision)
    bCollision = CheckBoundCollisions(x, y);

  return bCollision;
}

// ̭nʧ@禡, ⪺ʧ@
void TSprite::Move()
{
  if (FPostToDead) return; // wgnOǳƳQRF
  if (!FActive) return; // iʧ@

  // Cʧ@e𦸼
  if (FMoveDelay != 0) {
    // C FMoveDelay ~u Move @
    FMoveDelayCount++;
    if (FMoveDelayCount != FMoveDelay) return;
    FMoveDelayCount = 0; // ks
  }

  if (FFrameMax > 0) { // Yʵe
    FFrameNo++; // Wʵes
    if (FFrameNo > FFrameMax) { // @ӽjF
      if (FAdvanceFrameMode == afWrap)
        FFrameNo = 0; // sA
      else {
        FFrameNo = FFrameMax;  // b̫@iʵe
        FActive = false; // ʧ@
      }
    }
  }

  if (FSpeed == 0) return; // Yt׬s, 

  // oثem
  int x = FX;
  int y = FY;

  // pU@Bm
  switch (FDirection) {
    case drUp: y -= FSpeed; break; // W
    case drDown: y += FSpeed; break; // U
    case drLeft: x -= FSpeed; break; // 
    case drRight: x += FSpeed; break; // k
  }

  // ˬdҦI
  if (CheckCollisions(x, y))
    FCollisionCount++; //  줰FF...
  else
    FCollisionCount = 0; // 򳣨S, sIpƾks

  // sm
  FX = x;
  FY = y;

  // sΪxΰϰ
  FRect = ObjectRect;
  OffsetRect(&FRect, FX, FY);
}

// HNܤV
void TSprite::RandomDirection()
{
  Direction = (TDirection)random((int)drRight + 1);
}

// HN\m
void TSprite::RandomPosition()
{
  FX = random(WORLD_WIDTH - ObjectWidth);
  FY = random(WORLD_HEIGHT - ObjectHeight);
  
  FRect = ObjectRect;
  OffsetRect(&FRect, FX, FY);
}

// Pt@m
void TSprite::CenterWith(TSprite& ASprite)
{
  FX = ASprite.FX + ASprite.ObjectWidth / 2 - ObjectWidth / 2;
  FY = ASprite.FY + ASprite.ObjectHeight / 2 - ObjectHeight / 2;
  
  FRect = ObjectRect;
  OffsetRect(&FRect, FX, FY);
}

// ϤI (X, Y)
void TSprite::CenterBy(int x, int y)
{
  FX = x - ObjectWidth / 2;
  FY = y - ObjectHeight / 2;

  FRect = ObjectRect;
  OffsetRect(&FRect, FX, FY);
}

int TSprite::GetTile_X()
{
  return(FX + ObjectWidth / 2) / TILE_WIDTH;
}

int TSprite::GetTile_Y()
{
  return (FY + ObjectHeight / 2) / TILE_HEIGHT;
}

// oe
int TSprite::GetObjectWidth()
{
  return FObjectRect.Right - FObjectRect.Left;
}

// oⰪ
int TSprite::GetObjectHeight()
{
  return FObjectRect.Bottom - FObjectRect.Top;
}

void TSprite::SetDirection(TDirection Value)
{
  if (Value != FDirection) {
    FCollisionCount = 0; // V, sIks
    FDirection = Value;
  }
}

// nOR
void TSprite::PostToDie(WPARAM Param)
{
  if (FPostToDead) return; // еnO

  // ǰe WM_DESTROY_OBJECT TΥ}, ѹCjBz
  // Window handle  0, ܶǰe current thread
  PostMessage(0, WM_DESTROY_OBJECT, WPARAM(this), Param);
  FPostToDead = true;
}

// TTank abstract class

TTank::TTank()
{
  FFrameMax = 1; // ZJSʵe
  
  ResetStatus(); // mZJA (ͩRO, t, )

  // N[J Tanks }C
  Tanks()->push_back(this);
}

TTank::~TTank()
{
  TSpriteArray& r_Tanks = *Tanks();
  
  // N Tanks }C (FڤZJ)
  for (TSpriteArray::iterator p = r_Tanks.begin(); p != r_Tanks.end(); p++)
    if (*p == this) {
      r_Tanks.erase(p);
      break;
    }
}

// oglu, Ǧ^ȬogXhlu
TBullet* TTank::FireBullet()
{
  // ثeluƭYwFWhe\Aog
  if (FBulletNum >= FMaxBulletNum) return NULL;

  TBullet* b = new TBullet(this); // إߤlu
  b->LoadBits(); // Jluϧ
  b->Direction = Direction; // luPZJPˤV

  b->CenterWith(*this); // Nluyг]wPZJm

  switch (Direction) { // ھڦiVӽվluy
    case drUp: b->PosY = Rect.Top; break;
    case drDown: b->PosY = Rect.Bottom - b->ObjectHeight; break;
    case drLeft: b->PosX = Rect.Left; break;
    case drRight: b->PosX = Rect.Right - b->ObjectWidth; break;
  }

  FBulletNum++; // WZJثelu
  return b;
}

bool TTank::CheckCollisions(int& x, int& y)
{
  bool bCollision = TSprite::CheckCollisions(x, y);

  // ˬdO_ZJ
  if (!bCollision && !Attr.Contains(saNoTankCollision)) {
    TSpriteArray TankCollisions;
    bCollision = CheckTankCollisions(x, y, TankCollisions);
  }

  return bCollision;
}

bool TTank::CheckTankCollisions(int& x, int& y, TSpriteArray& Collisions)
{
  // Is CheckSpriteCollisions koIZJC
  TSpriteArray& r_Tanks = *Tanks();
  bool bCollision = TSprite::CheckSpriteCollisions(r_Tanks, x, y, Collisions);

  if (bCollision) { // YPZJo͸I
    for (TSpriteArray::iterator p = Collisions.begin(); p != Collisions.end(); ) {
      TSprite* &T = *p;

      if (T->Attr.Contains(saNoTankCollision))
        Collisions.erase(p);
      else
        p++;
    }

    bCollision = Collisions.size() > 0; // sMO_o͸I
    if (!bCollision) return false; // `ϸɨc, Sߤ]. ڥ⦳I..

    // ھڸI쪺ZJΦۤvVӽվy
    for (TSpriteArray::iterator p = Collisions.begin(); p != Collisions.end(); p++) {
      TSprite* &T = *p;

      switch (Direction) {
        case drUp: y = T->Rect.Bottom; break; // ۤvb誺U
        case drDown: y = T->Rect.Top - ObjectHeight; break;
        case drLeft: x = T->Rect.Right; break; // ۤvb誺k
        case drRight: x = T->Rect.Left - ObjectWidth; break;
      }
    }
  }
  return bCollision;
}

// mZJA (ͩRO, t, )
void TTank::ResetStatus()
{
  TSprite::ResetStatus();
  
  FSuperMode = false; // `Ҧ

  FHP = 1; // Q@N

  FMaxBulletNum = 1;
  FBulletNum = 0;
  FBulletBlowRange = BULLET_DEFAULT_BLOW_RANGE;

  Attr = TSpriteAttr() << saAlignWithTerrItem; // ZJ㦳۰ʽվmݩ, _Ӥ
}

// TMyTank class

TMyTank::TMyTank()
{
  Direction = drUp; // V¤W
  FFileName = "mytank.bmp";
}

void TMyTank::ResetStatus()
{
  TTank::ResetStatus();

  Speed = MYTANK_DEFAULT_SPEED; // ]wڤZJt
}

bool TMyTank::CheckGemCollisions(int& x, int& y)
{
  if (!Gem) return false; // _sb, iI

  // ھڷsyЭpZJҦΪxΰϰ
  TRect SpriteRect = ObjectRect;
  OffsetRect(&SpriteRect, x, y);

  // ZJO_P_Ҧxΰϰ첣ͥ涰
  TRect R, R1 = Gem->Rect;
  if (IntersectRect(&R, &SpriteRect, &R1))
    return true; // Y_F

  return false;
}

bool TMyTank::CheckCollisions(int& x, int& y)
{
  bool bCollision = TTank::CheckCollisions(x, y);

  // pGY_ (vT bCollision) ...
  if (CheckGemCollisions(x, y)) {

    // ھ_, ǰeTCjBz, ΦۤvBz
    switch (Gem->GemKind) {
      case gkClock:
      PostMessage(0, WM_SPECIAL_CONDITION, TIMER_ID_GEM_CLOCK, 0);
      break;

      case gkHat:
      PostMessage(0, WM_SPECIAL_CONDITION, TIMER_ID_GEM_HAT, 0);
      break;

      case gkArrow:
      PostMessage(0, WM_SPECIAL_CONDITION, TIMER_ID_GEM_ARROW, 0);
      break;

      case gkStar: FMaxBulletNum = 10; // solu
      break;

      case gkBlow: FBulletBlowRange = 32; // Wjluz¤O
      break;

      case gkApple: Speed = 2 * MYTANK_DEFAULT_SPEED; // ⭿t
      break;
    }

    // N_R(QYF)
    delete Gem;
  }
  
  return bCollision;
}

// TStar class

// ĤZJXͮɪ{{ĪGO
TStar::TStar()
{
  Speed = 0; // 
  Attr = TSpriteAttr() << saUndirectionalBitmap; // SV
  FMoveDelay = 8; // CK frame ~ܤ@
  FAdvanceFrameMode = afStop; // ʵeN

  FFileName = "star.bmp";
  FFrameMax = 2; // Tiʵe
  FObjectRect = Classes::Rect(0, 0, 38, 32);
}

// TETank class

// ĤZJO,
TETank::TETank(int KindNo)
{
  FKindNo = KindNo;
  Direction = drDown; // V@ߴ¤U

  FFrameMax = 1; // iʵe (CӤV)

  // bTӥXͦaIܤ@ (W, W, kW)
  PosY = 0;
  switch (random(3)) {
    case 0: PosX = 0; break;
    case 1: PosX = (WORLD_WIDTH - ObjectWidth) / 2; break;
    case 2: PosX = WORLD_WIDTH - ObjectWidth; break;
  }

  // Q TStar Osy{{ĪG
  FStar = new TStar;
  FStar->LoadBits();
  FStar->CenterWith(*this);

  // ɦW㦳Wߩ
  FFileName = "etank" + IntToStr(FKindNo) + ".bmp";
  
  switch (FKindNo) {
    case 1:
    FScore = 1000;
    break;

    case 2:
    Speed = SPRITE_DEFAULT_SPEED * 2; // tׯSO
    FScore = 1200;
    break;

    case 3:
    FScore = 1500;
    break;

    case 4:
    Speed = SPRITE_DEFAULT_SPEED - 1; // yLC@I
    FScore = 2000;
    HP = 4; // |U~|
    break;

    case 5:
    HP = 10; //  10 U~|, Wr
    Speed = SPRITE_DEFAULT_SPEED / 2; // ܺC
    FScore = 10000;

    // SV, |PaΪΩZJۼ (]bŤ)
    Attr = TSpriteAttr() << saUndirectionalBitmap << saNoCellCollision << saNoTankCollision;
    OnAir = true; // bŤ

    FFrameMax = 0; // @i
    FObjectRect = Classes::Rect(0, 0, 86, 48); // nj"B"~~
    break;
  }
}

// ĤZJʤk
void TETank::Move()
{
  if (FStar) { // ٦bXͤ..
    FStar->Move(); // {{iU@

    if (!FStar->Active) { // Y{{ʵeqF..
      delete FStar; // R{{, U@ʼĤZJN|X{F..
      FStar = NULL;
    }
  } else {
    // ̶ü"i"oglu
    if (random(100) < PROBAB_ETANK_SHOT_BULLET * 100)
      FireBullet();

    // YsIƹLh, ζüƳ\i, hV
    if (CollisionCount > MAX_ETANK_COLLISION_COUNT ||
      random(100) < PROBAB_ETANK_RANDOM_TURN * 100)
      RandomDirection();

    TTank::Move(); // iU@
  }
}

void TETank::Draw(TCanvas* Canvas)
{
  if (!FStar)
    TTank::Draw(Canvas); // `øsʧ@
  else
    FStar->Draw(Canvas); // eX{{PP
}

TBullet* TETank::FireBullet()
{
  TBullet* b = TTank::FireBullet();

  if (FKindNo == 3) {
    // luSO
    if (b) b->Speed = BULLET_DEFAULT_SPEED * 2;
  }

  return b;
}
// TExplosion

TExplosion::TExplosion()
{
  FLeadToOver = false; // |ɭP Game Over
  
  FMoveDelay = 4; // C| frame ~ܤ@
  Speed = 0; // 
  Attr = TSpriteAttr() << saUndirectionalBitmap << saNoCellCollision << saNoTankCollision << saNoBulletCollision; // SV

  FAdvanceFrameMode = afStop; // ʵeN
}

// TSmallExplosion

TSmallExplosion::TSmallExplosion()
{
  FFileName = "smexp.bmp";
  FFrameMax = 4; // i
}

// TBigExplosion

TBigExplosion::TBigExplosion()
{
  FFileName = "bigexp.bmp";
  FFrameMax = 5; // i
  FObjectRect = Classes::Rect(0, 0, 64, 64);
}

// TBullet class

TBullet::TBullet(TTank* ATank)
{
  FTank = ATank; // OogZJ
  FExplosion = NULL;
  FTankBulletNum_Dropped = false;
  Speed = BULLET_DEFAULT_SPEED;

  FFileName = "bullet.bmp";
  FFrameMax = 0; // u@i
  FObjectRect = Classes::Rect(0, 0, 16, 16);

  // N[J Bullets }C
  Bullets()->push_back(this);
}

TBullet::~TBullet()
{
  TSpriteArray& r_Bullets = *Bullets();

  // N Bullets }C
  for (TSpriteArray::iterator p = r_Bullets.begin(); p != r_Bullets.end(); p++)
    if (*p == this) {
      r_Bullets.erase(p);
      break;
    }

//  DebugStrFmt("Destroy: Bullets Num = %d", OPENARRAY(TVarRec, ((int)Bullets()->size())));

  // pGezF, N| FBulletNum, YSz (pPluۼ), NnbB
  if (FTank && !FTankBulletNum_Dropped)
    FTank->FBulletNum--; // ZJluƥ

  // Rz
  if (FExplosion) delete FExplosion;
}

void TBullet::FireExplosion(const std::type_info& typeinfo)
{
  if (FExplosion) return; // wgsbzF..

 // z
  if (typeinfo == typeid(TSmallExplosion))
    FExplosion = new TSmallExplosion;
  else
    FExplosion = new TBigExplosion;

  FExplosion->LoadBits();
  FExplosion->CenterWith(*this);

  if (FTank) { // YogZJsb
    FTank->FBulletNum--; // ZJluƥ
    FTankBulletNum_Dropped = true;
  }
}

void TBullet::LocateExplosion(TRect R)
{
  if (!FExplosion) return;

  switch (Direction) {
    case drUp: FExplosion->CenterBy(PosX, R.Bottom); break;
    case drDown: FExplosion->CenterBy(PosX, R.Top); break;
    case drLeft: FExplosion->CenterBy(R.Right, PosY); break;
    case drRight: FExplosion->CenterBy(R.Left, PosY); break;
  }
}

bool TBullet::CheckCollisions(int& x, int& y)
{
  bool bCollision = TSprite::CheckCollisions(x, y);

  // ˬdO_ZJ
  if (!bCollision && !Attr.Contains(saNoTankCollision)) {
    TSpriteArray TankCollisions;
    bCollision = CheckTankCollisions(x, y, TankCollisions);
  }

  // ˬdO_I䥦lu
  if (!bCollision) {
    TSpriteArray BulletCollisions;
    bCollision = CheckBulletCollisions(x, y, BulletCollisions);
  }

  return bCollision;
}

bool TBullet::CheckBoundCollisions(int& x, int& y)
{
  bool bCollision = TSprite::CheckBoundCollisions(x, y);

  // YluW
  if (bCollision && !FExplosion) {
    FireExplosion(typeid(TSmallExplosion)); // ޵oz
    FExplosion->CenterBy(PosX + ObjectWidth / 2, PosY + ObjectHeight / 2);
  }

  return bCollision;
}

bool TBullet::CheckCellCollisions(int Layer, int& x, int& y, TCellArray& Collisions)
{
  int OrgX = x;
  int OrgY = y;
  bool bCollision = TSprite::CheckCellCollisions(Layer, x, y, Collisions);

  // pGluPY  pass  cell IF
  if (bCollision && !FExplosion) {
    // MuluiHqL Cells
    for (TCellArray::iterator p = Collisions.begin(); p != Collisions.end(); ) {
      TCell* &C = *p;

      if (C->TileAttr.Contains(taBulletCanPass)) // luiHL
        Collisions.erase(p);
      else
        p++;
    }

    bCollision = Collisions.size() > 0;
    if (!bCollision) {      // `ϸɨc, Sߤ]. ڥ⦳I..
      x = OrgX; // _Ӧm
      y = OrgY;
      return false;
    }

    FireExplosion(typeid(TSmallExplosion)); // ޵oz
    FExplosion->CenterBy(x + ObjectWidth / 2, y + ObjectHeight / 2);

    // zvTd
    TRect ExpRect = ObjectRect;
    OffsetRect(&ExpRect, x, y);

    if (FTank)
      InflateRect(&ExpRect, FTank->FBulletBlowRange, FTank->FBulletBlowRange);
    else
      InflateRect(&ExpRect, BULLET_DEFAULT_BLOW_RANGE, BULLET_DEFAULT_BLOW_RANGE);

    for (TCellArray::iterator p = Collisions.begin(); p != Collisions.end(); p++) {
      TCell* &C = *p;

      if (C->TileAttr.Contains(taCanBreak)) { // O_iH} cell
        LocateExplosion(C->Rect);
        C->BreakBy(ExpRect);
      } else if (C->TileAttr.Contains(taFlag)) { // pGOxXQ
        C->TileNo = 25; // aXl
        FExplosion->FLeadToOver = true; // z|ϱoC
      }
    }
  }

  return bCollision;
}

bool TBullet::CheckTankCollisions(int& x, int& y, TSpriteArray& Collisions)
{
  if (!FTank) return false;  // ۤvZJIF, luLkˤH..
  if (FExplosion) return false; // wgzF, LkAˤH

  TSpriteArray& r_Tanks = *Tanks();
  bool bCollision = TSprite::CheckSpriteCollisions(r_Tanks, x, y, Collisions);

  if (bCollision) {
    for (TSpriteArray::iterator p = Collisions.begin(); p != Collisions.end(); p++) {
      TTank* T = dynamic_cast<TTank*>(*p);

      // luLkog, @wnOĤ}~⥴
      if (T != FTank && !T->Attr.Contains(saNoBulletCollision) &&
        (dynamic_cast<TMyTank*>(T) && dynamic_cast<TETank*>(FTank)) ||
        (dynamic_cast<TETank*>(T) && dynamic_cast<TMyTank*>(FTank))) {
        
        PlaySound(AnsiString(AppDir + "bomb.wav").c_str(), 0, SND_FILENAME | SND_ASYNC);
        
        FireExplosion(typeid(TBigExplosion)); // ޵ojz

        if (T->FSuperMode) continue; // LļҦ, ZJS

        T->FHP--; // ͩRO@
        if (T->FHP == 0) { // Q쪺ZJ¼F..
          FExplosion->CenterWith(*T); // ɦbZJz
          T->PostToDie(T->FScore);
        } else
          LocateExplosion(T->Rect); // S, bZJtz

        break; // @Ӥlu̦h@ZJ
      }
    }
  }

  return bCollision;
}

bool TBullet::CheckBulletCollisions(int& x, int& y, TSpriteArray& Collisions)
{
  if (!FTank) return false; // ۤvZJIF, luNS@
  if (FExplosion) return false; // wzF, Az

  TRect SpriteRect = ObjectRect;
  OffsetRect(&SpriteRect, x, y);

  TSpriteArray& r_Bullets = *Bullets();

  // luＲ|
  for (TSpriteArray::iterator p = r_Bullets.begin(); p != r_Bullets.end(); p++) {
    TBullet* B = dynamic_cast<TBullet*>(*p);
    TRect BulletRect = B->Rect;

    TRect R;
    if (this != B && B->FTank && B->Visible && !B->FExplosion &&
      ((dynamic_cast<TMyTank*>(B->FTank) && dynamic_cast<TETank*>(FTank)) ||
       (dynamic_cast<TETank*>(B->FTank) && dynamic_cast<TMyTank*>(FTank))) &&
      IntersectRect(&R, &SpriteRect, &BulletRect)) {

      // ѭѶ
      PostToDie();
      B->PostToDie();
    }
  }

  return false; // luIluI
}

void TBullet::Move()
{
  if (!FExplosion) // ٥z
    TSprite::Move();
  else {
    FExplosion->Move();

    if (!FExplosion->Active) {  // z
      if (FExplosion->FLeadToOver)
        PostMessage(0, WM_GAMEOVER, 0, 0);
      else
        FExplosion->Visible = false; // ð_

      // lu
      PostToDie();
    }
  }
}

void TBullet::Draw(TCanvas* Canvas)
{
  if (!FExplosion) // Y`i, heXlu
    TSprite::Draw(Canvas);
  else
    FExplosion->Draw(Canvas); // _heXz
}

// TGem class

TGem::TGem(TGemKind AGemKind)
{
  FGemKind = AGemKind;

  Speed = 0;
  Attr = TSpriteAttr() << saUndirectionalBitmap;

  FFrameMax = 0; // u@i
  // ھ_Ūϧ
  const AnsiString FileNames[6] =
    {"gemClock.bmp", "gemHat.bmp", "gemArrow.bmp", "gemStar.bmp", "gemBlow.bmp", "gemApple.bmp"};
  FFileName = FileNames[(int)FGemKind];

  Gem = this;
}

TGem::~TGem()
{
  Gem = NULL;
}

// ZJBz禡 (, øs, )

void MoveTanks()
{
  for (TSpriteArray::const_iterator p = Tanks()->begin(); p != Tanks()->end(); p++)
    (*p)->Move();
}

void DrawTanks(TCanvas* Canvas, bool OnAir)
{
  for (TSpriteArray::const_iterator p = Tanks()->begin(); p != Tanks()->end(); p++)
    if (OnAir == dynamic_cast<TTank*>(*p)->OnAir)
      (*p)->Draw(Canvas);
}

void FreeTanks()
{
  // u free ĤZJ@ !
  while (Tanks()->size() > 1) {
    TSprite* t = Tanks()->back();
    Tanks()->pop_back();
    delete t;
  }
}

// luBz禡 (, øs, )

void MoveBullets()
{
  for (TSpriteArray::const_iterator p = Bullets()->begin(); p != Bullets()->end(); p++)
    (*p)->Move();
}

void DrawBullets(TCanvas* Canvas) // øsҦlu
{
  for (TSpriteArray::const_iterator p = Bullets()->begin(); p != Bullets()->end(); p++)
    (*p)->Draw(Canvas);
}

void FreeBullets()
{
  while (Bullets()->size() > 0) {
    TSprite* b = Bullets()->back();
    Bullets()->pop_back();
    delete b;
  }
}

void FreeBulletsForTank(TTank* ATank)
{
  for (TSpriteArray::const_iterator p = Bullets()->begin(); p != Bullets()->end(); p++) {
    TBullet* b = dynamic_cast<TBullet*>(*p);
    if (b->Tank == ATank)
      b->Tank = NULL;
  }
}

TSpriteArray* Tanks()
{
  static TSpriteArray* _Tanks = NULL;

  if (_Tanks == NULL) {
    _Tanks = new TSpriteArray;
  }

  return _Tanks;
}

TSpriteArray* Bullets()
{
  static TSpriteArray* _Bullets = NULL;

  if (_Bullets == NULL) {
    _Bullets = new TSpriteArray;
  }

  return _Bullets;
}

